import { isFunction } from '../utils'
import { IDeserializer, IIterableAccess, IMapAccess, isVisitor, IVisitor } from './interface'

export class GenericSeed<T> {
  readonly visitor: IVisitor<T>

  constructor(visitor: IVisitor<T> = new Visitor()) {
    this.visitor = visitor
  }

  static deserialize<T, D extends IDeserializer>(deserializer: D, visitor: IVisitor<T> = new Visitor()): T {
    return deserializer.deserializeAny(visitor)
  }

  deserialize<D extends IDeserializer>(deserializer: D): T {
    return GenericSeed.deserialize(deserializer, this.visitor)
  }
}

export class Visitor<T> implements IVisitor<T> {
  private overrides?: Partial<IVisitor<T>>

  constructor(overrides?: Partial<IVisitor<T>>) {
    this.overrides = overrides
  }

  static from<T>(visitor: Partial<IVisitor<T>>): IVisitor<T> {
    if (visitor instanceof Visitor || isVisitor(visitor)) {
      return visitor
    }

    return new this(visitor)
  }

  visitBoolean(value: boolean): T {
    return this.overrides?.visitBoolean?.(value) ?? value as T
  }

  visitNumber(value: number): T {
    return this.overrides?.visitNumber?.(value) ?? value as T
  }

  visitBigInt(value: bigint): T {
    return this.overrides?.visitBigInt?.(value) ?? value as T
  }

  visitString(value: string): T {
    return this.overrides?.visitString?.(value) ?? value as T
  }

  visitSymbol(value: symbol): T {
    return this.overrides?.visitSymbol?.(value) ?? value as T
  }

  visitNull(): T {
    return this.overrides?.visitNull?.() ?? null as T
  }

  visitObject(access: IMapAccess): T {
    if (isFunction(this.overrides?.visitObject)) {
      return this.overrides?.visitObject(access)
    }

    const result = []

    for (const entry of access) {
      result.push(entry)
    }

    return Object.fromEntries(result)
  }

  visitIterable(access: IIterableAccess): T {
    if (isFunction(this.overrides?.visitIterable)) {
      return this.overrides?.visitIterable(access)
    }

    const result = []

    for (const element of access) {
      result.push(element)
    }

    return result as T
  }
}

